/* ACR2000 Kernel Driver
 * Copyright (C) 1998 George Lebl
 *
 * Licensed under GPL, Read file COPYING
 */
/*
 * This driver was done using the excellent Linux Device Drivers book
 * written by Alessandro Rubini as a guide.
 *
 * This driver implements a "protocol" you can use to talk to the
 * ACR2000 card for controlling steppers and servos, right now it's
 * using a device with major number of 127 (you can change that and
 * the addresses below). It should also work with acr8000 I'd guess.
 *
 * It's using buffers the size of a page for reads and writes and uses a
 * timer to poll for data and to write data, you can control how often it
 * polls with poll_ms and how much it can read/write on one timer with
 * in_one_poll
 *
 * v0.3 - added blocking reads and writes
 *	  added read/write buffers
 *	  added a few sanity checks
 *	  added select call
 *	  changed license to GPL
 *
 * v0.2 - made variables for configuration with insmod
 *        export no symbols
 *        took out debugging info
 *
 * v0.1 - Initial Release
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

#define ACR2000_DEFAULT_ADDR		0x300
#define ACR2000_DEFUALT_MAJOR_DEV_NUM	127

#define ACR2000_RANGE			0x004
#define ACR2000_STATUS_BYTE_OFFSET	0x002

#define ACR2000_VERSION			"0.3"


/*these are the changable parameters*/
int io = ACR2000_DEFAULT_ADDR;
int major_num = ACR2000_DEFUALT_MAJOR_DEV_NUM;
int ibuffer_len = PAGE_SIZE;
int obuffer_len = PAGE_SIZE;
int in_one_poll = 64; /*number of bytes to read or write on one timer*/
int poll_ms = 100;

#ifndef NULL
#define NULL ((void *)0)
#endif

static struct wait_queue *inq=NULL,*outq=NULL; /*read and write queue*/

static char *ibuffer=NULL;
static char *ibufp=NULL;
static int ibufplen=0;

static char *obuffer=NULL;
static char *obufp=NULL;
static int obufplen=0;

static struct timer_list acr2000_timer;

static int reading=0;
static int writing=0;

static int poll_jiffies = HZ/100; /*poll_ms converted to jiffies*/

static int
acr2000_detect(unsigned int port, unsigned int range)
{
	int err;

	if((err=check_region(port,range)) < 0)
		return err;

	request_region(port,range,"acr2000");

	return 0;
}

static void
acr2000_release(unsigned int port, unsigned int range)
{
	release_region(port,range);
}

static void
acr2000_poll_write(void)
{
	unsigned long flags;
	unsigned in,out;
	int size;

	for(size=0;size<in_one_poll && obufplen;size++) {
		out = *obufp;

		if((++obufp)>=obuffer+obuffer_len)
			obufp = obuffer;
		obufplen--;

		save_flags(flags);
		cli();

		in = inb_p(io + ACR2000_STATUS_BYTE_OFFSET);

		if(!(in & 0x80)) {
			restore_flags(flags);
			break;
		}
		outb_p(out,io+1);

		restore_flags(flags);
	}
	if(size)
		wake_up_interruptible(&outq);
}


static void
acr2000_poll_read(void)
{
	unsigned long flags;
	unsigned in;
	char *p;
	int size;

	for(size=0;size<in_one_poll;size++) {
		save_flags(flags);
		cli();

		in = inb_p(io + ACR2000_STATUS_BYTE_OFFSET);

		if(!(in & 0x20)) {
			restore_flags(flags);
			break;
		}
		in = inb_p(io);
		restore_flags(flags);

		p = ibufp+ibufplen;

		while(p>=ibuffer+ibuffer_len)
			p-=ibuffer_len;

		*p = in;

		if((++ibufplen)>ibuffer_len) {
			ibufplen--;
			if((++ibufp)==(ibuffer+ibuffer_len))
				ibufp = ibuffer;
		}
	}
	if(size)
		wake_up_interruptible(&inq);
}
	
static void
acr2000_timer_timeout(unsigned long data)
{
	acr2000_poll_read();
	acr2000_poll_write();

	init_timer(&acr2000_timer);
	acr2000_timer.expires = jiffies + poll_jiffies;
	acr2000_timer.function = acr2000_timer_timeout;
	add_timer(&acr2000_timer);
}

static int
acr2000_read(struct inode *inode, struct file *file, char *s, int size)
{
	int i;
	int ns;

	if(size<=0)
		return 0;

	/*try polling frist, we might get lucky*/
	acr2000_poll_read();

	/* block until there is something in the buffer*/
	while(ibufplen==0) {
		if(file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		interruptible_sleep_on(&inq);
		if(current->signal & ~current->blocked)
			return -ERESTARTSYS;
	}

	if(size>ibufplen)
		size = ibufplen;

	ns = size;

	/*we need to wrap around*/
	if(ibufp+ns > (ibuffer+ibuffer_len)) {
		i = (ibuffer+ibuffer_len)-ibufp;
		memcpy_tofs(s,ibufp,i);
		s+=i;
		ibufp=ibuffer;
		ibufplen-=i;
		ns-=i;
	}

	/*copy the remaining bytes*/
	memcpy_tofs(s,ibufp,ns);
	ibufp+=ns;
	ibufplen-=ns;
	if(ibufp == ibuffer+ibuffer_len)
		ibufp = ibuffer;

	return size;
}

static int
acr2000_write(struct inode *inode, struct file *file, const char *s, int size)
{
	int ns;
	int i;
	char *endp;

	if(size<=0)
		return 0;

	/*try polling frist, we might get lucky and clear some buffer*/
	acr2000_poll_write();

	/* block until there is some space in the buffer*/
	while(obuffer_len == obufplen) {
		if(file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		interruptible_sleep_on(&outq);
		if(current->signal & ~current->blocked)
			return -ERESTARTSYS;
	}

	if(size>(obuffer_len-obufplen))
		size=obuffer_len-obufplen;

	ns = size;

	endp = obufp+obufplen;
	while(endp>=(obuffer+obuffer_len))
		endp-=obuffer_len;


	/*we need to wrap around*/
	if(endp+ns > (obuffer+obuffer_len)) {
		i = (obuffer+obuffer_len)-endp;
		memcpy_fromfs(endp,s,i);
		s+=i;
		endp=obuffer;
		obufplen+=i;
		ns-=i;
	}

	/*copy the remaining bytes*/
	memcpy_fromfs(endp,s,ns);
	obufplen+=ns;

	/*try to write it now*/
	acr2000_poll_write();

	return size;
}

static int
acr2000_open(struct inode *inode, struct file *file)
{
	if((file->f_mode & FMODE_READ) && reading)
		return -EBUSY;
	if((file->f_mode & FMODE_WRITE) && writing)
		return -EBUSY;
	reading = (file->f_mode & FMODE_READ)!=0;
	writing = (file->f_mode & FMODE_WRITE)!=0;

	MOD_INC_USE_COUNT;
	return 0;
}

static void
acr2000_close(struct inode *inode, struct file *file)
{
	if(file->f_mode & FMODE_READ)
		reading = 0;
	if(file->f_mode & FMODE_WRITE)
		writing = 0;

	MOD_DEC_USE_COUNT;
}

static int
acr2000_select (struct inode *inode, struct file *filp,
		int mode, select_table *table)
{
	if(mode == SEL_IN) {
		if(ibufplen)
			return 1;
		select_wait(&inq, table);
		return 0;
	} else if(mode == SEL_OUT) {
		if(obuffer_len > obufplen)
			return 1;
		select_wait(&outq, table);
		return 0;
	}
	return 0;
}

struct file_operations acr2000_fops = {
	NULL, /*lseek*/
	acr2000_read,
	acr2000_write,
	NULL, /*readdir*/
 	acr2000_select,
	NULL, /*ioctl*/
	NULL, /*mmap*/
	acr2000_open,
	acr2000_close,
	NULL
	
};


static int
acr2000_register_dev(void)
{
	int err;

	if((err=register_chrdev(major_num,"acr2000",&acr2000_fops)) < 0)
		return err;
	return 0;
}

static void
acr2000_unregister_dev(void)
{
	unregister_chrdev(major_num,"acr2000");
}

int
init_module(void)
{
	int err;

	printk(KERN_INFO"ACR2000 Driver revision "ACR2000_VERSION", "
	       "(c) 1998 George Lebl, Licensed under the GPL\n");

	register_symtab(NULL);
	poll_jiffies = ((HZ*poll_ms)/1000);
	ibuffer = vmalloc(ibuffer_len);
	if(!ibuffer) return -ENOMEM;
	ibufp = ibuffer;
	ibufplen = 0;
		
	obuffer = vmalloc(obuffer_len);
	if(!obuffer) {
		err = -ENOMEM;
		goto free_ibuffer;
	}
	obufp = obuffer;
	obufplen = 0;

	err = acr2000_detect(io,ACR2000_RANGE);
	if(err) goto free_obuffer;

	err = acr2000_register_dev();
	if(err) goto unreg_port;

	init_timer(&acr2000_timer);
	acr2000_timer.expires = jiffies + poll_jiffies;
	acr2000_timer.function = acr2000_timer_timeout;
	add_timer(&acr2000_timer);

	return 0;

	unreg_port: acr2000_release(io,ACR2000_RANGE);
	free_obuffer: vfree(obuffer);
	free_ibuffer: vfree(ibuffer);
	return err;
}

void
cleanup_module(void)
{
	del_timer(&acr2000_timer);
	acr2000_unregister_dev();
	acr2000_release(io,ACR2000_RANGE);
	vfree(ibuffer);
	vfree(obuffer);
}
